/*
 *      Copyright Richard Gorton, 2007-2011 All rights reserved
 *
 *      Permission is granted to use, duplicate, and
 *      publish this text and program as long as it
 *      includes this entire comment block and limited
 *      rights reference.
 */

#ifndef _PATHPROF_H
#define _PATHPROF_H

/*
 * For paths lengths < VLPATHLIM, we track individually.
 * Aggregate paths >= VLPATHLIM as a single entity.
 * Pick VLPATHLIM as 2048 which ought to be nearly impossible to encounter
 * In practice, path lengths rarely exceed 170 Bbs.  In all of SPEC CPU 2006,
 * there are only about 1000 unique paths with lengths greater than 170;
 * the longest path yet seen is 208Bbs in length.
 *
 * Investigative (aka loose end) curiousity - what are the relevant dynamic
 * frequencies of these long paths?  Where do they occur: in some infrequently
 * used (at least in the SPEC benchmarks) chunk of library code?  Application code?
 */

#define VLPATHLIM       2048
typedef struct PSTATS {
    uint64_t       totalInstructions;
    uint64_t       totalBbs;
    uint64_t       totalPaths;
    uint64_t       pathDistribution[VLPATHLIM];
    uint64_t       VLPaths;
    uint64_t       uniquePaths;
    uint64_t       uniquePathLenSum;
    unsigned int        maxPathBbLen;
    unsigned int        maxPathInsnLen;
    uint64_t       highestTripCount;
    uint64_t       strataBuffers;
    } PSTATS;

typedef struct BB_BLOB {
    void*       BbAddress;
    } BB_BLOB;

// PathIDs start at 1
// Need to know this for playback
#define FIRST_PATH_NUM  1
/*
 * Using SPEC CPU 2006 (base compile, using gcc 4.1.2): the largest cases are:
 * Total Instructions:  7393068165213   ref/calculix
 * Total Bbs:            552781903248   ref/calculix
 * Total Paths:          256403809900   ref/caclulix
 * Max Insns per path:           2432   train/tonto, ref/tonto
 * Max Bbs per path:              345   test/wrf, train/wrf, ref/wrf
 * Mean path length (Bbs):         21   ref/perlbench
 * Unique Paths:              5158297   ref/gobmk
 * Mean #Bbs/unique path:          76   ref/sjeng
 *
 * No paths yet seen exceed 2048 Bbs in length.
 *
 * Other interesting tidbits discovered
 * Highest continuous tripcount on a path: 27328510 for ref/mcf
 *
 * Use these values to design an viable hash function.  Hopefully
 * one which is moderately efficient.
 *
 */
typedef uint64_t   addrHash_T;

typedef struct PATH_QUEUE_ELT {
    unsigned int        pathID;         // 445.gobmk ref: 0x4eb599
    unsigned short      pathLen;        // 481.wrf train: 0x159
    unsigned short      pathInsns;      // 465.tonto train: 0x980
    addrHash_T          addrHash;       // Here for performance
    uint64_t       tripCount;
    struct BB_BLOB      dataBlob[1];    // Actually an indeterminate length array
    } PATH_QUEUE_ELT;


// Useful for extracting data from the db
typedef struct MAX_PATH_ELT {
    unsigned int        pathID;
    unsigned short      pathLen;
    unsigned short      pathInsns;
    addrHash_T          addrHash;
    uint64_t       tripCount;
    struct BB_BLOB      dataBlob[VLPATHLIM];
    } MAX_PATH_ELT;

/*
 * We initially create a PATH_QUEUE_ELT to insert in the hash table.
 * tripCount is used to improve performance for the case of tight loops
 * Our raw description of the application behavior is simply a sequence
 * of pathIDs.
 * Speculation (proven) : intuitively, there will be many cycles of paths, which we
 * can then treat as higher layer metacycles.  Geology has the notion of various
 * strata, so call these layers upon layers "stratum" ids.
 * These will consume the bulk of the space, so we need to be careful in their
 * construction.
 *
 * Better algorithmic compression might be attainable via an algorithm like
 * Larus' sequitur grammar, rather than a brute force approach.
 * The catch being that playback starting at an arbitrary point does become
 * relatively expensive with such an approach - one has to play back the entire
 * stored grammar just to arrive at the relevant starting point.  With the
 * brute force mechanism, a binary search approach can be used.
 *
 * The elephant in the room turns out to be the number of paths.  gobmk is
 * a brutal consumer of memory.  Lots of fine tuning is used to try to avoid
 * the data collection from swapping.
 *
 */

typedef struct STRATUM_STATS {
    uint64_t       longestLayer;
    uint64_t       VLLayers;
    uint64_t       strataDistrib[VLPATHLIM];
    uint64_t       strataMatches[VLPATHLIM];
    uint64_t       numDBEntries;
    uint64_t       consecutive;
    uint64_t       layer0cycles;   // equivalent of totalPaths
    uint64_t       uniquelayer0cycles;
    uint64_t       totalSize;
    uint64_t       totalCompressedSize;
    } STRATUM_STATS;

/*
 * The stratum tracking is very similar to path tracking, except we don't know
 * or care about the number of instructions, basic blocks, or how "hot" it is.
 * The minimum length of a stratum cycle is 2 separate paths, and this means our
 * initial dataBlob can be size 2.  A length of a cycle == 1 means path P1 with
 * T1 trips is followed by path P1 T2 trips.  But occurance is represented as 
 * path P1 with (T1 + T2) trips.
 */

// Grrr - wasting the extra bits on a combination of pathID + tripCount is
// annoying.  (circa 2007, when the out-of-pocket-cost to build an 8Gb machine
// to collect all of the data in main memory without swapping was significant)
// And I _did_ pay for the entire system & operational costs out of my own pocket.
// Now (2011), 8Gb is cheap, and it is not even particularly expensive to construct
// a 16Gb system.
typedef struct FAT_PATH_BLOB {
    unsigned int        pathID;
    uint64_t            tripCount;
    } FAT_PATH_BLOB;

/*
 * Use SMALL_PATH_BLOBs to save space.  We need a consistency check in the
 * path generator at the point at which we pass the blobs via shared memory.
 * If we get an error from that point, more work will be needed to cope with
 * memory crunch problem.
 */
typedef struct SMALL_PATH_BLOB {
    unsigned int        pathID;
    unsigned int        tripCount;
    } SMALL_PATH_BLOB;

typedef union SPB_U {
    struct SMALL_PATH_BLOB      Blob;
    uint64_t                    l_Blob;
    } SPB_U;

typedef struct STRATUM_ELT {
    unsigned int        stratumID;      // unique - represents a cycle of paths
    unsigned short      stratumLen;     // in # of paths; max seen so far is 287
    unsigned short      addHash;        // sum of path ids
    uint64_t            tripCount;
    union SPB_U         dataBlob[2];    // indeterminate length array
    } STRATUM_ELT;

/*
 * Some statistics about the strata
 * Number of unique strata:         2862441 ref/bzip2
 * Longest strata (in paths):           444 test/wrf train/wrf ref/wrf
 * Highest continuous tripcount: 5121578643 ref/sphinx
 */

// Used to create relevant PureDB entries
typedef struct datum {
    char*       dptr;
    size_t      dsize;
    } datum;
/*
 * Descriptions of the databases and their key/value pairs
 * The ...pathprof_db database:
 *      Key             Value
 *      ---------       ------------------------
 *      NumPaths        Number unique paths (4 byte int)
 *      PathStats       PSTATS structure contents
 *      PathId:<int>    PATH_QUEUE_ELT with corresponding pathID range 1..<max>
 *      0x<Bbaddr>      Compressed string of the Disassembly of the
 *                      basic block at <bbaddr>
 *
 * the Bbaddr key is expressed as generated by sprintf (buf, "%p", <bbaddr>)
 *
 * The strata databases:
 */

// Put the revision number into the DB to ensure some sort of mechanism to
// make incompatible changes
#define KEY_VERSION     1
#define KEY_VERNO       "VerNum"
#define KEY_NUMPATHS    "NumPaths"
#define KEY_PATHSTATS   "PathStats"
#define KEY_PATHID      "PathID"
#define KEY_NUMSTRATA   "NumStrata"
#define KEY_STRATUMSTATS "StratumStats"
#define KEY_STRATUMID   "StratumID"

#ifdef __cplusplus
extern "C" {
#endif
void PathInit(char *dbFnam, char *baseName);
void StoreDisAsm (void *BbAddr, char *decode);
void ExtendPath (void *BbAddr, unsigned int numBbInsns);
void PathFini();
void StrataInit(char *dbFnam);
void ExtendStrataLayer (union SPB_U blob);
void StrataFini();
#ifdef __cplusplus
};
#endif

extern FILE*    LogF;

// Diagnostics...
#endif  // _PATHPROF_H
