/*
 *      Copyright Richard Gorton, 2007-2011 All rights reserved
 *
 *      Permission is granted to use, duplicate, and
 *      publish this text and program as long as it
 *      includes this entire comment block and limited
 *      rights reference.
 */
/*
 *      Author: Rick Gorton     richard.gorton@amd.com
 *
 *      path consumer/analyzer code, part of a DWPPP: Dynamic Whole Program
 *      Path Profiler.
 *      At tool creation, it was not possible to go from an address to a BBL in PIN,
 *      the entire disassembly had to be carried around for the run.  This was both
 *      a hassle, and fortuituous: 450.soplex had an ascii (disassembly) footprint
 *      of 13Gb, which was a lot at that time, mostly in terms of the hardware cost
 *      for an individual to foot the bill for.  This also blew sdbm out of the water.
 *      gdbm turned out to do poorly, but because the content is basically write-once,
 *      compression + PureDB works well.  Switched to PureDB in Aug, 2007.
 *      The fortuitous part: the net result is that the ascii disassembly is now
 *      stored as part of the resulting content, and is thus preserved.  That is,
 *      due to the fact it was a challenge, the disassembly is guaranteed to be preserved.
 *
 *  Notes/Comments:
 *      A stand-alone disassembler (unattached to PIN) might be the way to go,
 *      long term.  Either that, or have a back-end (separate) tool do the decode,
 *      and merely dump the Bb addresses (and maybe lengths) with this tool.
 *
 *      Aug, 2007 - dump GDBM for PureDB
 *              http://pureftpd.sourceforge.net/puredb/
 *      PureDB can handle up to 4Gb of content.  This should not present a problem for the
 *      compressed disassembly + path information, except for truly enormous applications
 *      For actual content of 1x, gdbm seems to use a bit more than 2x (in two files) to
 *      hold the data, presumably to support deletion & modification.
 *      And it seems to try to keep the entire thing in memory, which makes my (as a developer)
 *      life difficult when trying to do a SPEC CPU 2006 reference run (gobmk ends up swapping
 *      some on a system with 8Gb of memory), even with heroic efforts.
 */

#include <sys/types.h>
#include <fcntl.h>
#include <limits.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>

#ifdef _WIN32
// An attempt at porting the tools to windows.  A dead end, as there is zero value-add
#define INLINE          __inline
#define PATH_MAX        256
#include "windows/stdint.h"
#include <search.h>
// Have to have search.h for qsort.  Utterly stupid!
#define snprintf        _snprintf
#define unlink        _unlink
#else
#define INLINE          inline
#include <stdint.h>
#include <unistd.h>
#include <zlib.h>
#endif

#include "puredb/puredb_read.h"
#include "puredb/puredb_write.h"

static PureDBW  pathDB_pureW;
static PureDB   pathDB_pureR;
static char     dbFName[PATH_MAX];

#include "pathprof.h"
#include "servercomm.h"

#ifdef WPPP_DEBUG
static int      DBGCheckPathHash = 0;
#endif
static void ProcessPath (struct PATH_QUEUE_ELT* aPath);

#ifdef BUILD_WPPP
static int finalFlush = 0;
#endif
struct PSTATS pathStats;

/*
 * Path structs and variables
 * QC_CACHESIZE - the size of a 'quick cache' to catch tight loops prior to passing off the work
 * to the heavier-weight processing of ProcessPath()
 * In path profiler mode, the full cache will be used in order to boost performance.
 * In wppp (whole program path profiler) mode, only one entry can be used to preserve
 * complete accuracy (and the count stratum fields will get exceeded frequently).
 *
 * SPECIAL ALGORITHMIC OPTIMIZATION for WPPP mode, when QC_CACHESIZE == 1
 *      When the QC_CACHESIZE == 1, we can change when we allocate new entries for the hash
 *      table, and we never have to do a free in ProcessPath.  This will save some compute
 *      cycles, which is CRITICAL to do for the WPPP.
 *      To do this, use a single maximally sized oversized 'path'.
 */

#ifdef BUILD_WPPP
static struct PATH_QUEUE_ELT* s_prevPath = NULL; 
static uint64_t         s_phonyBlob[VLPATHLIM + 4];
#define QC_CACHESIZE    1
#else
#define QC_CACHESIZE    8
#endif

static struct BB_BLOB   s_tmpPath [VLPATHLIM + 1];
static addrHash_T       s_addrHash = 0;
static struct PATH_QUEUE_ELT*   s_quickCache[QC_CACHESIZE];

/*
 * Big hash table - 256K wide.
 * An earlier version was 64K wide, but was way too small (in part due to a bug).
 * but it does turn out that some applications do have a huge number of unique paths.
 * A good example of this is 445.gobmk from SPEC CPU 2006.  One of the training runs
 * has 1.83E6 paths (and I suspect the ref run has even more - it has not yet run).
 */
typedef struct PSTORE_ENT {
    struct PATH_QUEUE_ELT*      item;
    struct PSTORE_ENT*          next;
    } PSTORE_ENT;

#define PHSIZE  (256*1024)
static struct PSTORE_ENT**      PHash;

static uint64_t         s_curPathBbLen = 0;
static uint64_t         s_curPathInsnLen = 0;
static unsigned short   s_pathInsns = 0;

#define PATH_HASH_CHECK_FREQ    (uint64_t) 10e6;
static uint64_t         PathHashCheck = PATH_HASH_CHECK_FREQ;
#define PATH_HASH_SORT_FREQ     (uint64_t) 100e6;
static uint64_t         PathHashSort = PATH_HASH_SORT_FREQ;


// Forward declarations
static void dumpSinglePath (struct PATH_QUEUE_ELT* curPath);

/*
 * Sort by descending order of: tripcount of the path
 * This is to speed up accesses to the paths in spite of possible phase changes.
 * We're passed PSTORE_ENT addresses
 */
int PHChainCmp (const void *item1, const void* item2)
{
    struct PSTORE_ENT*  p1_pp;
    struct PSTORE_ENT*  p2_pp;
    struct PATH_QUEUE_ELT*      path1_p;
    struct PATH_QUEUE_ELT*      path2_p;
    int         res;

    p1_pp = (struct PSTORE_ENT*) item1;
    p2_pp = (struct PSTORE_ENT*) item2;
    path1_p = p1_pp->item;
    path2_p = p2_pp->item;

    if (path2_p->tripCount > path1_p->tripCount)
        res = 1;
    else if (path2_p->tripCount < path1_p->tripCount)
        res = -1;
    else
        res = 0;

    return res;
}

/*
 * Sort by descending order of: tripcount * number insns in the path
 * We're passed PSTORE_ENT addresses
 * This is to determine which paths are hottest.
 */
static int PathCmp (const void *item1, const void* item2)
{
    struct PSTORE_ENT*  p1_pp;
    struct PSTORE_ENT*  p2_pp;
    struct PATH_QUEUE_ELT*      path1_p;
    struct PATH_QUEUE_ELT*      path2_p;
    int         res;
    uint64_t    qty1;
    uint64_t    qty2;

    p1_pp = (struct PSTORE_ENT*) item1;
    p2_pp = (struct PSTORE_ENT*) item2;
    path1_p = p1_pp->item;
    path2_p = p2_pp->item;

    qty2 = (path2_p->tripCount * path2_p->pathInsns);
    qty1 = (path1_p->tripCount * path1_p->pathInsns);

    if (qty2 > qty1)
        res = 1;
    else if (qty2 < qty1)
        res = -1;
    else
        res = 0;

    return res;
}

#define DEPTH_TRIGGER   1
#define TRIPS_TRIGGER   100000
typedef struct SORTREC {
    struct PSTORE_ENT*  pse;
    uint64_t            trips;
    } SORTREC;

static void RestructurePathHash()
{
    uint64_t    maxTrips;
    int         i;

    for (i = 0; i < PHSIZE; i++) {
        struct PSTORE_ENT*      base;
        int depth;

        depth = 0;
        maxTrips = 0;
        base = PHash[i];
        while (base != NULL) {
            uint64_t    curTrips;

            curTrips = base->item->tripCount;
            if (curTrips > maxTrips) {
                maxTrips = curTrips;
            }
            depth++;
            base = base->next;
            
        }

        /*
         * Do not reorder for trivial chains.  Either too short a length, or all
         * members having trivial trip counts.
         * 23-Oct-2007
         * For those chains which we do sort, use a customized sort.
         * Unfortunately, this is a necessity, as using qsort inside of PIN
         * causes some of the SPEC2006 (gobmk/test on dniwog dataset) benchmarks
         * to fault for no apparent reason.  But only on a particular value, etc.
         * The fault seems to stem from a corruption of the data pointed to by
         * one of the pointers.  It goes from being valid, to the memory address
         * no longer being valid.  This may be a PIN problem, but it has to be
         * worked around
         */
        if ((depth > DEPTH_TRIGGER) && (maxTrips > TRIPS_TRIGGER)) {
            struct SORTREC*     toBeSorted;
            struct SORTREC      newrec;
            struct PSTORE_ENT*  walker;
            int                 j;
            int                 k;
            int                 l;

                // Fill a linear table and do a descending value insertion sort
            base = PHash[i];
            toBeSorted = (struct SORTREC*) calloc (depth, sizeof(*toBeSorted));
            for (j = 0; j < depth; j++) {
                newrec.pse = base;
                newrec.trips = base->item->tripCount;
                base = base->next;
                // Simply tack trivial entries onto the end, and we are done
                if (newrec.trips < TRIPS_TRIGGER) {
                    toBeSorted[j] = newrec;
                    continue;
                }

                // Find correct spot to insert, and shift the rest
                k = 0;
                while ((k < j) && (toBeSorted[k].trips > newrec.trips)) {
                    k++;
                }

                // At the correct location: shift remainder
                l = j;
                while ((l > k) && (l > 0)) {
                    toBeSorted[l] = toBeSorted[l - 1];
                    l--;
                }

                // Insert the new
                toBeSorted[k] = newrec;
            }

                // then unpack the table back into the hash chain
            PHash[i] = toBeSorted[0].pse;
            walker = toBeSorted[0].pse;
            for (j = 0; j < depth - 1; j++) {
                walker->next = toBeSorted[j + 1].pse;
                walker = walker->next;
            }
            walker->next = NULL;        // Force null termination

                // Free the table
            free (toBeSorted);
        }       // Was the hash slot worth sorting?
    }   // Each hash slot
}

/*
 * create an entry (if necessary) in the db of the form:
 * Key = <bbaddr> - expressed with sprintf (buf, "%p", <bbaddr>)
 * Value = compressed disassembly (by PIN) of said Bb
 *
 * MAJOR ASSUMPTION: the number of instructions for a BbAddr NEVER, EVER changes
 * but this may not be true, based upon PIN documentation.
 */
void StoreDisAsm (void *BbAddr, char *disassembly)
{
    datum       Key;
    datum       Value;
    int         KeyLen;
    char        KeyStr[128];
    uint64_t    rawDisLen;
    uint64_t    compDisLen;
    char        compressBuf[1024*1024];
    int         compres;
    int         rc;

        // Build the key
    snprintf (KeyStr, sizeof(KeyStr), "%p", BbAddr);
    KeyLen = strlen (KeyStr);
    Key.dptr = KeyStr;
    Key.dsize = KeyLen;

        // Compress the input: length + 1 for trailing null; 9 is max compression
    rawDisLen = strlen (disassembly) + 1;
    compDisLen = sizeof(compressBuf);
#ifdef _WIN32
    memcpy (compressBuf, disassembly, rawDisLen);
    compressBuf[rawDisLen] = 0x0;
    compDisLen = rawDisLen;
#else
    compres = compress2 (compressBuf, &compDisLen, disassembly, rawDisLen, 9);
#endif

    Value.dptr = compressBuf;
    Value.dsize = compDisLen;
    rc = puredbw_add (&pathDB_pureW, Key.dptr, Key.dsize, Value.dptr, Value.dsize);
    if (rc != 0) {
        fprintf (LogF, "Bad puredbw_add: %d\n", rc);
    }
}

static INLINE struct PATH_QUEUE_ELT* AllocateElt (uint64_t pLen)
{
    int pathSize;
    struct PATH_QUEUE_ELT* newElt;

        // pLen is the current length of the path; subtract one
        // for the first element being in the structure already
    pathSize = sizeof(struct PATH_QUEUE_ELT) + ((pLen - 1) * sizeof(struct BB_BLOB));
    newElt = (struct PATH_QUEUE_ELT*) malloc (pathSize);

    return newElt;
}

static INLINE void FillElt (struct PATH_QUEUE_ELT* Elt, uint64_t pLen, uint64_t addV, uint64_t pathInsns)
{
    int j;

    Elt->pathLen = pLen;
    Elt->tripCount = 1;
    Elt->addrHash = addV;
    Elt->pathInsns = pathInsns;
    for (j = 0; j < s_curPathBbLen; j++) {
        Elt->dataBlob[j] = s_tmpPath[j];
    }
}

static void AddToQuickCache (struct PATH_QUEUE_ELT* qc_item)
{
    int i;

        // Transfer tail to processPath
    if (s_quickCache[QC_CACHESIZE - 1] != NULL) {
        ProcessPath (s_quickCache[QC_CACHESIZE - 1]);
    }
    for (i = QC_CACHESIZE - 1; i > 0; i--) {
        s_quickCache[i] = s_quickCache[i - 1];
    }
    s_quickCache[0] = qc_item;
}

/*
 * This is actually moot, since it will only occur when the program exits, so there
 * is only going to be a use count of one.
 */
static void FlushPathQueue ()
{
#ifdef BUILD_WPPP
    ProcessPath (s_prevPath);
    FillElt (s_prevPath, s_curPathBbLen, s_addrHash, s_curPathInsnLen);
    finalFlush = 1;
    ProcessPath (s_prevPath);
#else
    int i;
    struct PATH_QUEUE_ELT*      newPath;

        // Empty the quickcache first
    for (i = QC_CACHESIZE - 1; i >= 0; i--) {
        ProcessPath (s_quickCache[i]);
        s_quickCache[i] = NULL;
    }
    newPath = AllocateElt (s_curPathBbLen);
    FillElt (newPath, s_curPathBbLen, s_addrHash, s_curPathInsnLen);
    ProcessPath (newPath);
#endif

    return;
}

static void CheckHashFn()
{
    struct PSTORE_ENT*  walker;
    int badHashfn;
    int i;

    badHashfn = 0;
    for (i = 0; i < PHSIZE; i++) {
        int     sum;

        walker = PHash[i];
        sum = 0;
        while (walker) {
            sum++;
            walker = walker->next;
        }

        // If we see a depth of more than about 5% that means we have a problem
        if (sum * 20 > pathStats.uniquePaths) {
            badHashfn = 1;
            break;
        }
    }   // Each slot

    if (badHashfn != 0) {
        fprintf (LogF, "*****************************\n");
        fprintf (LogF, "Poor Path Hash table distribution - consider reworking GenHashVal()\n"
                "Important info: (abnormally deep slots)\nSlot\t(Hex)\tDepth\n");
        for (i = 0; i < PHSIZE; i++) {
            int     sum;

            walker = PHash[i];
            sum = 0;
            while (walker) {
                sum++;
                walker = walker->next;
            }
                // Report anything deeper than 0.1%
            if ((sum * 1000) > pathStats.uniquePaths) {
                fprintf (LogF, "%d\t%x\t%d\n", i, i, sum);
            }
        }
        fprintf (LogF, "*****************************\n");
    } else {
        fprintf (LogF, "Info: Good Path hash distribution: (hash function works well)\n");
    }
}

#if BUILD_WPPP
static void WriteLongToPathDB (char *keystr, uint64_t val)
{
    const char* func = "WriteIntToPathDB";
    uint64_t    scratchValue;
    char        dbkey[2048];
    datum       Key;
    datum       Value;
    int         rc;

    snprintf (dbkey, sizeof(dbkey), "%s", keystr);
    scratchValue = val;
    Key.dptr = dbkey;
    Key.dsize = strlen (dbkey);
    Value.dptr = (char *) &scratchValue;
    Value.dsize = sizeof (scratchValue);
    rc = puredbw_add (&pathDB_pureW, Key.dptr, Key.dsize, Value.dptr, Value.dsize);
    if (rc != 0) {
        fprintf (LogF, "Bad puredbw_add in %s for key: [%s]: %d\n", func, keystr, rc);
    }
}

static void WriteStructToPathDB (char *keystr, char *valBuf, uint64_t valBufSize)
{
    const char* func = "WriteStructToPathDB";
    char        dbkey[2048];
    datum       Key;
    datum       Value;
    int         rc;

    snprintf (dbkey, sizeof(dbkey), "%s", keystr);
    Key.dptr = dbkey;
    Key.dsize = strlen (dbkey);
    Value.dptr = (char *) valBuf;
    Value.dsize = valBufSize;
    rc = puredbw_add (&pathDB_pureW, Key.dptr, Key.dsize, Value.dptr, Value.dsize);
    if (rc != 0) {
        fprintf (LogF, "Bad puredbw_add in %s for key: [%s]: %d\n", func, keystr, rc);
    }
}

static void WritePathsDB (struct PSTORE_ENT*  pTable)
{
    const char  *func="WritePathsDB";
    int     i;

        /*
         * Write the DB version so consumer tools can understand how to extract stuff
         * Number of paths - simply here as a matter of convenience - it could be extracted
         * by extracting the pathstats struct...
         */

    WriteLongToPathDB (KEY_VERNO, KEY_VERSION);     // for expandability
    WriteLongToPathDB (KEY_NUMPATHS, pathStats.uniquePaths);

        /*
         * The full pathstats structure
         */
    WriteStructToPathDB (KEY_PATHSTATS, (char*) &pathStats, sizeof(pathStats));

        /*
         * Write the individual pathid structures
         */
    for (i = 0; i < pathStats.uniquePaths; i++) {
        char        dbkey[2048];
        struct PATH_QUEUE_ELT       *curPath;
        int         curPathSize;

        curPath = pTable[i].item;
        curPathSize = sizeof (struct PATH_QUEUE_ELT) + (sizeof (struct BB_BLOB) * (curPath->pathLen - 1));
        snprintf (dbkey, sizeof(dbkey), "%s:%d", KEY_PATHID, curPath->pathID);
        WriteStructToPathDB (dbkey, (char*) curPath, curPathSize);
    }
}
#endif

static void ClosePathsDB()
{
    (void) puredbw_close (&pathDB_pureW);
    puredbw_free (&pathDB_pureW);
}

//  the database is assumed to be open
static void
dumpSinglePath (struct PATH_QUEUE_ELT* curPath)
{
    const char  *func = "dumpSinglePath";
    float       pathInsPercent;
    float       pathBbPercent;
    void*       dbReadp;
    uint64_t    dynInsCount;
    uint64_t    dynBbCount;
    int         rc;
    int         j;

    dynInsCount = curPath->tripCount * curPath->pathInsns;
    dynBbCount = curPath->tripCount * curPath->pathLen;
    pathInsPercent = (float) dynInsCount * (float) 1.000e2 / (float) pathStats.totalInstructions;
    pathBbPercent = (float) dynBbCount * (float) 1.000e2 / (float) pathStats.totalBbs;
        // Something broken on windows with fprintf - if done as a single fprintf,
        // Insns shows as a zero with PathLength getting Insns.
        // Probably related to uint64_t
    fprintf (LogF, "pathID:\t%d\t", curPath->pathID);
    fprintf (LogF, "Trips:\t%ld\t", curPath->tripCount);
    fprintf (LogF, "Insns: %d\t", curPath->pathInsns);
    fprintf (LogF, "PathLength: %d\t", curPath->pathLen);
    fprintf (LogF, "\nPath Instructions: %ld\t", dynInsCount);
    fprintf (LogF, "Percent of total execution: %f\n", pathInsPercent);
    fprintf (LogF, "\nPath Bbs: %ld\t", dynBbCount);
    fprintf (LogF, "Percent of total execution: %f\n", pathBbPercent);
    fprintf (LogF, "\nBbAddresses:\t");
    for (j = 0; j < curPath->pathLen; j++) {
            // Print 4 addresses wide
        fprintf (LogF, "%p\t", curPath->dataBlob[j].BbAddress);
        if (j%4 == 3) {
            fprintf (LogF, "\n\t\t");
        }
    }       // All the Bb Addresses
    fprintf (LogF, "\n\tPATH DISASSEMBLY:\n");
    for (j = 0; j < curPath->pathLen; j++) {
        char        KeyStr[128];
        datum       BbKey;
        datum       Value;
        char        uncompBuf[1024 * 1024];
        uint64_t    uncompLen;

        snprintf (KeyStr, sizeof(KeyStr), "%p", curPath->dataBlob[j].BbAddress);
        BbKey.dptr = KeyStr;
        BbKey.dsize = strlen (KeyStr);
        rc = puredb_find (&pathDB_pureR, BbKey.dptr, BbKey.dsize,
                              (off_t *) &Value.dptr, (size_t *) &Value.dsize);
        if (rc != 0) {
            fprintf (LogF, "puredb_find error in %s: %d\n", func, rc);
        } else {
            dbReadp = puredb_read (&pathDB_pureR, (off_t) Value.dptr, Value.dsize);
#ifndef _WIN32
            uncompLen = sizeof (uncompBuf);
            uncompress (uncompBuf, &uncompLen, (char *)dbReadp, Value.dsize);
#else
            snprintf (uncompBuf, sizeof(uncompBuf), "%s", dbReadp);
            uncompBuf[sizeof(uncompBuf)] = 0x0;
#endif
            puredb_read_free (dbReadp);
        }
        fprintf (LogF, "%s", uncompBuf);
    }
}

// Maximum number of Paths to report
#define DUMP_MAX        20
// Another summary stat report percentages for the hottest TOP_LIM paths
#define TOP_LIM         5

static void ReportHottestPaths (struct PSTORE_ENT*  pTable)
{
    const char* func = "ReportHottestPaths";
    int i;
    int rc;
    float       netInsnPercent;
    float       topInsnPercent;
    void*       dbReadp;

    rc = puredb_open (&pathDB_pureR, dbFName);  // opened solely for read
        // hottest DUMP_MAX paths...
    fprintf (LogF, "\nPath profile characteristics\n");
    netInsnPercent = 0.0e0;
    topInsnPercent = 0.0e0;

    fprintf (LogF, "\n----------------------------------------------------\n"
                "Hottest %d paths (sorted by tripcount * insns/path)\n", DUMP_MAX);
    for (i = 0; i < DUMP_MAX; i++) {
        int     j;
        uint64_t        dynInsCount;
        float   pathInsPercent;
        struct PATH_QUEUE_ELT   *curPath;

        fprintf (LogF, "\n------------------------- Hottest Path # %d -------------------------\n", i + 1);
        curPath = pTable[i].item;

        dumpSinglePath (curPath);

        // "heat" is really cycles rather than instruction count, but to a
        // first order of magnitude, instruction count is good enough.
        dynInsCount = curPath->tripCount * curPath->pathInsns;
        pathInsPercent = (float) dynInsCount * (float) 1.000e2 / (float) pathStats.totalInstructions;
        netInsnPercent += pathInsPercent;
        if (i < TOP_LIM) {
            topInsnPercent += pathInsPercent;
        }

    }
    (void) puredb_close (&pathDB_pureR);
    fprintf (LogF, "\n----------------------------------------------------\n");
    fprintf (LogF, "Instructions executed by top %d paths: %4.2f%%\n", TOP_LIM, topInsnPercent);
    fprintf (LogF, "Instructions executed by top %d paths: %4.2f%%\n", DUMP_MAX, netInsnPercent);
}

/*
 * Dump distribution of the path hash table
 * Assumes LogF is open.
 */
void PathFini()
{
    const char* func = "PathFini";
    struct PSTORE_ENT*  walker;
    int i;
    int tabIdx;
    struct PSTORE_ENT*  pTable;
    uint64_t            uniquePathLenMean;

    FlushPathQueue();
    CheckHashFn();

        // Allocate and build a table of the unique paths so we can sort & report
        // This will be an array of pointers to the path structs.

    pTable = (struct PSTORE_ENT*) calloc (pathStats.uniquePaths, sizeof(*pTable));
    tabIdx = 0;
    for (i = 0; i < PHSIZE; i++) {
        walker = PHash[i];
        while (walker) {
            memcpy ((void *)&pTable[tabIdx], (void*) walker, sizeof(*pTable));
            tabIdx++;
            walker = walker->next;
        }
    }

    if (tabIdx > pathStats.uniquePaths) {
        fprintf (stderr, "**** tabIdx screwup: %d\n", tabIdx);
        fprintf (LogF, "**** tabIdx screwup: %d\n", tabIdx);
    }

#if BUILD_WPPP
    WritePathsDB (pTable);
#endif

    ClosePathsDB();

    fprintf (LogF, "\nPath definition: a path is defined as a sequence of basic blocks\n"
                "which do not repeat.  If the Bb sequence seen is AABCDCBADD, the paths are:\n"
                "\tA\tABCD\tCBAD\tD\n\n");

    fprintf (LogF, "Info: Number paths longer than %ld Bbs: %ld\n", VLPATHLIM, pathStats.VLPaths);
    fprintf (LogF, "Info: Highest trip count going into ProcessPath: %ld\n", pathStats.highestTripCount);


    fprintf (LogF, "\nPath length distribution histogram\n");
    fprintf (LogF, "\tLength\tCount\n");
    for (i = 0; i < VLPATHLIM; i++) {
        if (pathStats.pathDistribution[i] != 0) {
            fprintf (LogF, "\t%ld\t%ld\n", i, pathStats.pathDistribution[i]);
        }
    }

        // Sort it all for reporting the hottest ones
    qsort ((void*) &pTable[0], tabIdx, sizeof (*pTable), PathCmp);
    ReportHottestPaths (pTable);

        // Dump stats in a form easily consumed by a spreadsheet
    uniquePathLenMean = pathStats.uniquePathLenSum / pathStats.uniquePaths;
    fprintf (LogF, "\nSpreadsheet usable stats\n");
    fprintf (LogF, "sshdr\tNumber\tNumber\tNumber\tMax\tMax\tMean\tNumber\tMean\tStrata\n");
    fprintf (LogF,
        "sshdr\tInsns\tBbs\tPaths\tInsns\tBbs\tBb\tUnique\tBbs\tBuffers\n"
        "sshdr\t\t\t\tper\tper\tPath\tPaths\tUnique\t\n"
        "sshdr\t\t\t\tPath\tPath\tLen\t\tPath\t\n");

    fprintf (LogF, "ssdata\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\n",
        pathStats.totalInstructions, pathStats.totalBbs, pathStats.totalPaths,
        (uint64_t) pathStats.maxPathInsnLen, (uint64_t) pathStats.maxPathBbLen,
        pathStats.totalBbs / pathStats.totalPaths,
        pathStats.uniquePaths, uniquePathLenMean,
        pathStats.strataBuffers);

#ifdef BUILD_WPPP
#else
        /*
         * Delete the paths db file in path profiler mode.  Keep it for wppp mode
         */
    unlink (dbFName);
#endif
}
/*
 * Batch up the ident/trip records as PATH_BLOBs and pass (via shared memory)
 * to the strata server.
 */
#ifdef BUILD_WPPP
static uint64_t         rawIdx = 0;
static struct SHARED_MEM*       s_sharedMem = NULL;
static struct SHARED_MEM        strataBuf;

static void PassToStrataServer (int pathIdent, uint64_t pTrips, int forcewrite)
{
    struct SMALL_PATH_BLOB    theData;

        /*
         * Do a sanity check here.  For memory compression purposes, PATH_BLOB
         * tripCount fields are int.  Otherwise holding the data for
         * apps such as SPEC2006/sjeng (ref run) does consume upwards of 7Gb
         * of memory in the strata.
         */

    if ((pTrips >> 31) > 0) {
        /*
         * Another possibility suggests itself here if this happens:
         * Issue multiples with values < 32 bits until we get a decent result
         * (recursively call this).  But that leaves us with the recognition
         * problem then the scanner when we see MAXINT trips exactly.
         * Using a convention whereby a trip count of zero really means MAX_INT
         * and there is followon data, otherwise the value is the true value
         * could make this work.
         */
        fprintf (LogF, "Error: huge number of path Trips (%ld) for path id: %d\n",
                pTrips, pathIdent);
    }
    theData.pathID = pathIdent;
    theData.tripCount = pTrips;

    if ((rawIdx + sizeof (theData) > DB_BUF_SIZE) || (forcewrite != 0)) {
        char    opFlags;

        opFlags = 0x0;
        if (pathStats.strataBuffers == 0) {
            opFlags |= OP_OPEN_APPEND;
        }
        if (forcewrite != 0) {
            opFlags |= OP_FLUSH_CLOSE;
        }
        pathStats.strataBuffers++;

        // Build up a message for the server to analyze/compress
        strataBuf.dataHdr.version = THIS_VERSION;
        strataBuf.dataHdr.operation = opFlags;
        strataBuf.dataHdr.flags = FLAG_COMPRESS_9;
        strataBuf.dataHdr.payloadSize = rawIdx;
        strataBuf.dataHdr.startInsnEpoch = pathStats.totalInstructions;
        strataBuf.dataHdr.startPathEpoch = pathStats.totalPaths;
        strataBuf.dataHdr.startStrataEpoch = 0;

        // Send the message to the server, once server is available
        WaitUntilReadyToFill (s_sharedMem);
        memcpy ((void*) s_sharedMem, (void*) &strataBuf, sizeof(strataBuf));
        MarkDataFilled (s_sharedMem);

        // reset for further use
        rawIdx = 0;
    }

        // copy new info into buffer
    memcpy ((void*) &strataBuf.rawDBBuf[rawIdx], (void*) &theData, sizeof(theData));
    rawIdx += sizeof(theData);

}
#endif

/*
 * This is a key performance routine.  Pointer chasing down the chains
 * is really slow when the chains get deep.  With the original algorithm, running
 * astar with lake.cfg (training?) took 12 hours.
 * Original algorithm: (BB[0].addr & 0xff00) | (xorhash & 0xf0) | (pathlen & 0xf)
 *              aaaa aaaa xxxx pppp
 * Adjustment 1: hmmer bombesin.hmm still chugging along at 13 hours
 * That algorithm: (ilenval & 0xf << 12) | ((addrHash (all Bbaddrs) & 0xff) >> 8)
 *              | (pathlen & 0xf << 8) 
 *              iiii pppp hhhh hhhh
 * Perhaps widening the table by another 2 bits (4x to 256K slots) is the way to go.
 * With overlapped values: ilen at the top; add in the addrhash of the address & 0x3fff
 * and add in the pathlen
 *      ii iiii 0000 0000 0000
 * Add  00 00hh hhhh hhhh hhhh
 * Add  pathlen (always < 2048)
 *
 * Postscript: widening does help a lot, but at the cost of extra memory.  Maybe
 * when it becomes possible to affordably build 16Gb systems with reasonable DDR speeds
 * 16Gb @ DDR2/667 is a non-starter.
 */
static unsigned int GenHashVal (struct PATH_QUEUE_ELT* aPath)
{
    unsigned int        val;
    unsigned int        addVal;
    unsigned int        iLenVal;
    unsigned int        BbLenVal;

    iLenVal = aPath->pathInsns & 0x003f;
    iLenVal = iLenVal << 12;
    addVal = aPath->addrHash & 0xffff;
    BbLenVal = aPath->pathLen;

    val = iLenVal + addVal + BbLenVal;
        // And be sure to retruncate
    val = val & 0x3ffff;

    if (val >= PHSIZE) {
        // Report the screwup, but be defensive and keep going
        fprintf (LogF, "GenHashVal consistency error - %d bigger than table size: %d\n",
                val, PHSIZE);
        val = PHSIZE - 1;
    }

    return val;
}

/*
 * save away this path - if we find it already exists, then update the count
 * of the path, and discard this new copy.
 * Note that this is the routine which owns the responsibility for freeing
 * the path vectors.
 * This is a pretty good place to hand off the data to another process
 * or thread for further analysis.
 */
static void ProcessPath (struct PATH_QUEUE_ELT* aPath)
{
    struct PSTORE_ENT*  walker;
    struct PSTORE_ENT*  newH;
    unsigned int        hIdx;
    uint64_t            pTrips;
    int                 i;
    int                 pathIdent;

    pTrips = aPath->tripCount;
    pathIdent = 0;
    if (pTrips > pathStats.highestTripCount) {
        pathStats.highestTripCount = pTrips;
    }
    hIdx = GenHashVal (aPath);
    walker = PHash[hIdx];

    while (walker != NULL) {
        if ((walker->item->dataBlob[0].BbAddress == aPath->dataBlob[0].BbAddress)
                && (walker->item->addrHash == aPath->addrHash)
                && (walker->item->pathInsns == aPath->pathInsns)
                && (walker->item->pathLen == aPath->pathLen)) {
                // Need to check for full match, but can start at index 1
            for (i = 1; i < aPath->pathLen; i++) {
                if (walker->item->dataBlob[i].BbAddress != aPath->dataBlob[i].BbAddress) {
                    break;
                }
            }
                // If a complete match, bump count on existing, discard new
            if (i == aPath->pathLen) {
                pathIdent = walker->item->pathID;
                walker->item->tripCount += pTrips;
#ifdef BUILD_WPPP
#else
                free (aPath);
#endif

                goto done;
            }
        }       // Partial match
        walker = walker->next;
    }   // walking the chain

        // Got here, means no pre-existing item.  Push onto the FRONT(!) of the
        // table assuming temporal locality will help with performance.
        // Also means we don't have to track a 'prev' pointer.
#ifdef BUILD_WPPP

        /*
         * Performance improvement in WPPP mode - only allocate a new structure when
         * we have to.  Which means we do not need to do a free() in the common case.
         */
    if (walker == NULL) {
        struct PATH_QUEUE_ELT*  newPath;
        int     j;

        newPath = AllocateElt (aPath->pathLen);
        newPath->pathLen = aPath->pathLen;
        newPath->tripCount = aPath->tripCount;
        newPath->addrHash = aPath->addrHash;
        newPath->pathInsns = aPath->pathInsns;
        for (j = 0; j < aPath->pathLen; j++) {
            newPath->dataBlob[j] = aPath->dataBlob[j];
        }
        pathStats.uniquePaths++;
        newPath->pathID = pathStats.uniquePaths;
        pathIdent = pathStats.uniquePaths;
        pathStats.uniquePathLenSum += newPath->pathLen;
        newH = (struct PSTORE_ENT*) malloc (sizeof (*newH));
        newH->item = newPath;
        newH->next = PHash[hIdx];
        PHash[hIdx] = newH;
        goto done;
    }
#else
    if (walker == NULL) {
        pathStats.uniquePaths++;
        aPath->pathID = pathStats.uniquePaths;
        pathIdent = pathStats.uniquePaths;
        pathStats.uniquePathLenSum += aPath->pathLen;
        newH = (struct PSTORE_ENT*) malloc (sizeof (*newH));
        newH->item = aPath;
        newH->next = PHash[hIdx];
        PHash[hIdx] = newH;
        goto done;
    }
#endif

done: ;

#ifdef BUILD_WPPP
    PassToStrataServer (pathIdent, pTrips, finalFlush);
#endif

        /*
         * Much like the reordering of the hash table and hash function checking
         * done in strata.c, periodically check the hash table to see if the
         * hash function is broken;
         * periodically reorder the entries in the hash table such that if there
         * are drastic phase changes, we put the more likely ones to be seen up front.
         */

#ifdef WPPP_DEBUG
    if (DBGCheckPathHash != 0) {
        PathHashCheck--;
        if (PathHashCheck <= 0) {
            PathHashCheck = PATH_HASH_CHECK_FREQ;
            CheckHashFn();
        }
    }
#endif

        // Re-order the hash slots.  Less frequently than checking of the hash function
        // Guess 1e9 paths.
    PathHashSort--;
    if (PathHashSort <= 0) {
        PathHashSort = PATH_HASH_SORT_FREQ;
        RestructurePathHash();
    }

    return;
}

void PathInit (char *dbFileName, char *baseName)
{
    const char* func = "InitPath";
    struct PATH_QUEUE_ELT       checkme;
    char        idxFName[PATH_MAX];
    char        datFName[PATH_MAX];
    int         rc;

    if (sizeof (s_addrHash) != sizeof (checkme.addrHash)) {
        /*
         * If they do not match, the unique paths will not be calculated correctly,
         * and consequently, the hash table will grow rapidly (which will slow things
         * down a lot)
         */
        fprintf (LogF, "Serious problem - local hash size in %s is mismatched - aborting\n", func);
        exit (1);
    }

    memset ((void*) &pathStats, 0, sizeof(pathStats));
    PHash = (struct PSTORE_ENT**) calloc (PHSIZE, sizeof (*PHash));

    snprintf (idxFName, sizeof (idxFName), "%s.p_tmpidx", dbFileName);
    snprintf (datFName, sizeof (datFName), "%s.p_tmpdat", dbFileName);
    snprintf (dbFName, sizeof (dbFName), "%s.p_dbp", dbFileName);

    rc = puredbw_open (&pathDB_pureW, idxFName, datFName, dbFName);
    if (rc != 0) {
        fprintf (LogF, "Bad open of paths puredb: %d - aborting\n", rc);
        exit (1);
    }

#ifdef BUILD_WPPP
        // Attach to the compression server shared memory region
    s_sharedMem = OpenMMapFile ("wppp_PIN_tool", CLIENT_OPEN_MMAPFILE, STRATA_MMAP_FILE);
    if (s_sharedMem == NULL) {
        fprintf (LogF, "wppp PIN tool: Bad open of %s: %d - aborting\n",
                STRATA_MMAP_FILE, rc);
        exit (1);
    }
    strcpy (strataBuf.dataHdr.dataFileBaseName, baseName);
#endif

}

/*
 * The path builder routine.  When a cycle of Bb addresses is discovered, 
 * check to see if this new path has been been recently contructed.  This will be
 * in the quickcache.  For the wppp, the size of the quickcache is always 1, so it
 * will be the immediately previously built one.  If we see a match, simply bump the
 * trip count and reset the various counters/tables.  If not, we have to construct
 * a path structure, and process it (it might already exist in the hash table).
 */
void ExtendPath (void *BbAddr, unsigned int numBbInsns)
{
    int         i;
    int         qcIdx;
    struct PATH_QUEUE_ELT*      newPath;

    pathStats.totalInstructions += numBbInsns;
    pathStats.totalBbs++;

    for (i = 0; i < VLPATHLIM; i++) {
        if (s_tmpPath[i].BbAddress == NULL) {
                // Not on the path so far, extend
            s_tmpPath[i].BbAddress = BbAddr;
            s_tmpPath[i+1].BbAddress = NULL;    // This is why s_tmpPath is of size VLPATHLIM + 1
            s_curPathBbLen = i + 1;
                // Doing this saves a large memset when starting a new path
            s_curPathInsnLen += numBbInsns;
            s_addrHash += (uint64_t) BbAddr;
            goto done;
        } else if (s_tmpPath[i].BbAddress == BbAddr) {
            int mismatch = 0;
                /*
                 * Path: adjust stats, queue up the path for the consumer thread.
                 * As a memory savings, check to see if this new path is exactly the
                 * same as the one just created.  If so, simply increment the trip count
                 * of this one, and do not create a new one.
                 */

            pathStats.totalPaths++;
            if (s_curPathBbLen > pathStats.maxPathBbLen) {
                pathStats.maxPathBbLen = s_curPathBbLen;
            }
            if (s_curPathInsnLen > pathStats.maxPathInsnLen) {
                pathStats.maxPathInsnLen = s_curPathInsnLen;
            }

            pathStats.pathDistribution[s_curPathBbLen]++;

#ifdef BUILD_WPPP
            if (s_prevPath == NULL) {
                s_prevPath = (struct PATH_QUEUE_ELT *) s_phonyBlob;
                FillElt (s_prevPath, s_curPathBbLen, s_addrHash, s_curPathInsnLen);
                goto reset;
            }

                // Check for immediate match
            if ((s_prevPath->addrHash == s_addrHash) && (s_prevPath->pathLen == s_curPathBbLen)) {
                int j;

                for (j = 0; j < s_curPathBbLen; j++) {
                    if (s_prevPath->dataBlob[j].BbAddress != s_tmpPath[j].BbAddress) {
                        ProcessPath (s_prevPath);
                        FillElt (s_prevPath, s_curPathBbLen, s_addrHash, s_curPathInsnLen);
                        goto reset;
                    }
                }
                // Complete match if we got here.  Simply bump trip count
                s_prevPath->tripCount++;
                goto reset;
            }

                // If we got here, not a match
            ProcessPath (s_prevPath);
            FillElt (s_prevPath, s_curPathBbLen, s_addrHash, s_curPathInsnLen);
            goto reset;

#else
                // Not WPPP mode, use a tunable-sized cache
                /*
                 * Check s_quickCache[] to see if we've got a an entry already,
                 * without doing the heavy-weight ProcessPath() work.
                 */
            for (qcIdx = 0; qcIdx < QC_CACHESIZE; qcIdx++) {
                if (s_quickCache[qcIdx] == NULL) {
                    newPath = AllocateElt (s_curPathBbLen);
                    FillElt (newPath, s_curPathBbLen, s_addrHash, s_curPathInsnLen);
                    AddToQuickCache (newPath);
                    goto reset;
                }       // empty (not yet filled) cache

                // Check for a match of current cache slot vs this new path.
                if ((s_quickCache[qcIdx]->addrHash == s_addrHash)
                        && (s_quickCache[qcIdx]->pathLen == s_curPathBbLen)) {
                    int j;

                    for (j = 0; j < s_curPathBbLen; j++) {
                        if (s_quickCache[qcIdx]->dataBlob[j].BbAddress != s_tmpPath[j].BbAddress) {
                            mismatch = 1;
                            break;
                        }
                    }   // Check all
                    if (mismatch == 0) {
                        // Complete match, simply increment, reset for next path
                        s_quickCache[qcIdx]->tripCount++;
                        goto reset;
                    }
                }       // partial match
            }   // Check each entry in s_quickCache

                // Different than previously cached items - add to cache.
            newPath = AllocateElt (s_curPathBbLen);
            FillElt (newPath, s_curPathBbLen, s_addrHash, s_curPathInsnLen);
            AddToQuickCache (newPath);
#endif

reset:
                // Reset things for new path
            s_tmpPath[0].BbAddress = BbAddr;
            s_tmpPath[1].BbAddress = NULL;
            s_curPathBbLen = 1;
            s_curPathInsnLen = numBbInsns;
            s_addrHash = (uint64_t) BbAddr;

            goto done;
        }       // Got a Bbaddr match, which means we have a new path
    }   // Walk up to the limit of s_tmpPath
        /*
         * Not found, and we are at the temp table end.  A very long path.
         * Reset as if we are starting a new path.  Note that this smears the
         * statistics because we are now
         * starting to look at a shortened path.
         * NOTE: if this happens a lot, it means that there is a really obnoxious case
         * and VPATHLIM probably needs to be doubled.
         */

    if (i == VLPATHLIM) {
        pathStats.VLPaths++;
        s_tmpPath[0].BbAddress = NULL;
        s_curPathBbLen = 1;
        s_curPathInsnLen = 0;
        s_pathInsns = 0;
        s_addrHash = 0;
        goto done;
    }

        // Consistency error.  We never should have gotten here.
    fprintf (LogF, "ExtendPath error - very bad problem: Bb == %p, i == %ld\n", BbAddr, i);

done:
    return;
}
